import torch as tr

# Mapping of nucleotide symbols
# R	Guanine / Adenine (purine)
# Y	Cytosine / Uracil (pyrimidine)
# K	Guanine / Uracil
# M	Adenine / Cytosine
# S	Guanine / Cytosine
# W	Adenine / Uracil
# B	Guanine / Uracil / Cytosine
# D	Guanine / Adenine / Uracil
# H	Adenine / Cytosine / Uracil
# V	Guanine / Cytosine / Adenine
# N	Adenine / Guanine / Cytosine / Uracil
NT_DICT = {
    "R": ["G", "A"],
    "Y": ["C", "U"],
    "K": ["G", "U"],
    "M": ["A", "C"],
    "S": ["G", "C"],
    "W": ["A", "U"],
    "B": ["G", "U", "C"],
    "D": ["G", "A", "U"],
    "H": ["A", "C", "U"],
    "V": ["G", "C", "A"],
    "N": ["G", "A", "C", "U"],
}

VOCABULARY = ["A", "C", "G", "U"]


class OneHotEmbedding:
    def __init__(self):
        self.pad_token = "-"
        self.vocabulary = VOCABULARY
        self.emb_size = len(self.vocabulary)

    def seq2emb(self, seq, pad_token="-"):
        """One-hot representation of seq nt in vocabulary.  Emb is CxL
        Other nt are mapped as shared activations.
        """
        seq = seq.upper().replace("T", "U")  # convert to RNA
        emb_size = len(VOCABULARY)
        emb = tr.zeros((emb_size, len(seq)), dtype=tr.float)

        for k, nt in enumerate(seq):
            if nt == pad_token:
                continue
            if nt in VOCABULARY:
                emb[VOCABULARY.index(nt), k] = 1
            elif nt in NT_DICT:
                v = 1 / len(NT_DICT[nt])
                ind = [VOCABULARY.index(n) for n in NT_DICT[nt]]
                emb[ind, k] = v
            else:
                raise ValueError(f"Unrecognized nucleotide {nt}")

        return emb



import json
import os
import random
import numpy as np
import torch as tr
from datetime import datetime
import pandas as pd
import shutil

from torch.utils.data import DataLoader
from sincfold.dataset import SeqDataset, pad_batch
from sincfold.model import sincfold
from sincfold.embeddings import NT_DICT
from sincfold.utils import write_ct, validate_file, ct2dot
from sincfold.parser import parser
from sincfold.utils import dot2png, ct2svg

__version__ = "0.12"

def main():
    
    args = parser()
    
    if not args.no_cache and args.command == "train":
        cache_path = "cache/"
    else:
        cache_path = None

    config= {"device": args.d, "batch_size": args.batch,  "use_restrictions": False, 
             "valid_split": 0.1, "max_len": args.max_length, "verbose": not args.quiet, "cache_path": cache_path}
    
    if "max_epochs" in args:
        config["max_epochs"] = args.max_epochs

    if args.config is not None:
        config.update(json.load(open(args.config)))

    if config["cache_path"] is not None:
        shutil.rmtree(cache_path, ignore_errors=True)
        os.makedirs(cache_path)

    # Reproducibility
    tr.manual_seed(42)
    random.seed(42)
    np.random.seed(42)

    if args.command == "train": 
        train(args.train_file, config, args.out_path,  args.valid_file, args.j)

    if args.command == "test":
        test(args.test_file, args.model_weights, args.output_file, config, args.j)

    if args.command == "pred":
        pred(args.pred_file, args.sequence_name, args.model_weights, args.output_file, config, args.j, args.draw, args.draw_resolution)    
        
def train(train_file, config={}, out_path=None, valid_file=None, nworkers=2, verbose=True):
    
    
    if out_path is None:
        out_path = f"results_{str(datetime.today()).replace(' ', '-')}/"
    else:
        out_path = out_path

    if verbose:
        print("Working on", out_path)

    if "cache_path" not in config:
        config["cache_path"] = "cache/"
    
    if not os.path.isdir(out_path):
        os.makedirs(out_path)
    else:
        raise ValueError(f"Output path {out_path} already exists")

    if valid_file is not None:
        train_file = train_file
        valid_file = valid_file
    else:
        data = pd.read_csv(train_file)
        valid_split = config["valid_split"] if "valid_split" in config else 0.1
        train_file = os.path.join(out_path, "train.csv")
        valid_file = os.path.join(out_path, "valid.csv")

        val_data = data.sample(frac = valid_split)
        val_data.to_csv(valid_file, index=False)
        data.drop(val_data.index).to_csv(train_file, index=False)
        
    batch_size = config["batch_size"] if "batch_size" in config else 4
    train_loader = DataLoader(
        SeqDataset(train_file, **config),
        batch_size=batch_size,
        shuffle=True,
        num_workers=nworkers,
        collate_fn=pad_batch,
    )
    valid_loader = DataLoader(
        SeqDataset(valid_file, **config),
        batch_size=batch_size,
        shuffle=False,
        num_workers=nworkers,
        collate_fn=pad_batch,
    )

    net = sincfold(**config)
    
    best_f1, patience_counter = -1, 0
    if verbose:
        print("Start training...")
    max_epochs = config["max_epochs"] if "max_epochs" in config else 1000
    for epoch in range(max_epochs):
        train_metrics = net.fit(train_loader)

        val_metrics = net.test(valid_loader)

        if val_metrics["f1"] > best_f1:
            best_f1 = val_metrics["f1"]
            tr.save(net.state_dict(), os.path.join(out_path, "weights.pmt"))
            patience_counter = 0
        else:
            patience_counter += 1
            if patience_counter > 30:
                break
        msg = (
            f"epoch {epoch}:"
            + " ".join([f"train_{k} {v:.3f}" for k, v in train_metrics.items()])
            + " "
            + " ".join([f"val_{k} {v:.3f}" for k, v in val_metrics.items()])
        )
        
        if verbose:
            print(msg)
        with open(os.path.join(out_path, "train.txt"), "a") as f: 
            f.write(msg + "\n")
            f.flush()
            
    # remove temporal files           
    shutil.rmtree(config["cache_path"], ignore_errors=True)
    
    tmp_file = os.path.join(out_path, "train.csv")
    if os.path.exists(tmp_file):
        os.remove(tmp_file)
    tmp_file = os.path.join(out_path, "valid.csv")
    if os.path.exists(tmp_file):
        os.remove(tmp_file)
    
def test(test_file, model_weights=None, output_file=None, config={}, nworkers=2, verbose=True):
    test_file = test_file
    test_file = validate_file(test_file)
    if verbose not in config:
        config["verbose"] = verbose

    test_loader = DataLoader(
        SeqDataset(test_file, **config),
        batch_size=config["batch_size"] if "batch_size" in config else 4,
        shuffle=False,
        num_workers=nworkers,
        collate_fn=pad_batch,
    )

    if model_weights is not None:
        net = sincfold(weights=model_weights, **config)
    else:
        net = sincfold(pretrained=True, **config)
    
    if verbose:
        print(f"Start test of {test_file}")        
    test_metrics = net.test(test_loader)
    summary = " ".join([f"test_{k} {v:.3f}" for k, v in test_metrics.items()])
    if output_file is not None:
        with open(output_file, "w") as f:
            f.write(summary+"\n")
    if verbose:
        print(summary)

def pred(pred_input, sequence_id='pred_id', model_weights=None, out_path=None, config={}, nworkers=2, draw=False, draw_resolution=10, verbose=True):
    
    if out_path is None:
        output_format = "text"
    else:
        _, ext = os.path.splitext(out_path)
        if ext == "":
            if os.path.isdir(out_path):
                raise ValueError(f"Output path {out_path} already exists")
            os.makedirs(out_path)
            output_format = "ct"
        elif ext != ".csv":
            raise ValueError(f"Output path must be a .csv file or a folder, not {ext}")
        else:
            output_format = "csv"

    file_input = os.path.isfile(pred_input)
    if file_input:
        pred_file = validate_file(pred_input)
    else:
        pred_input = pred_input.upper().strip()
        nt_set = set([i for item  in list(NT_DICT.values()) for i in item] + list(NT_DICT.keys()))
        if set(pred_input).issubset(nt_set):
            pred_file = f"{sequence_id}.csv"
            with open(pred_file, "w") as f:
                f.write("id,sequence\n")
                f.write(f"{sequence_id},{pred_input}\n")
            
        else:
            raise ValueError(f"Input have invalid inputs, nucleotides should be picked from: {nt_set}")
    pred_loader = DataLoader(
        SeqDataset(pred_file, max_len=config["max_len"] if "max_len" in config else 512, for_prediction=True, use_restrictions=config["use_restrictions"] if "use_restrictions" in config else False),
        batch_size=config["batch_size"] if "batch_size" in config else 4,
        shuffle=False,
        num_workers=nworkers,
        collate_fn=pad_batch,
    )
    
    if model_weights is not None:
        weights = model_weights
        net = sincfold(weights=weights, **config)
    else:
        net = sincfold(pretrained=True, **config)

    if verbose:        
        print(f"Start prediction of {pred_file}")
    predictions = net.pred(pred_loader)
    if draw:
        for i in range(len(predictions)):
            item = predictions.iloc[i]
            ctfile = "tmp.ct"
            write_ct(ctfile, item.id, item.sequence, item.base_pairs)
            dotbracket = ct2dot(ctfile)
            
            png_file = item.id +".png"
            if out_path is not None and os.path.isdir(out_path):
                png_file = os.path.join(out_path, png_file)
            if dotbracket:
                dot2png(png_file, item.sequence, dotbracket, resolution=draw_resolution)
            ct2svg("tmp.ct", png_file.replace(".png", ".svg"))

    if not file_input:
        os.remove(pred_file)

    if output_format == "text":
        for i in range(len(predictions)):
            item = predictions.iloc[i]
            ctfile = "tmp.ct"
            write_ct(ctfile, item.id, item.sequence, item.base_pairs)
            dotbracket = ct2dot(ctfile)
            print(item.id)
            print(item.sequence)
            print(dotbracket)
            print()
    elif output_format == "csv":
        predictions.to_csv(out_path, index=False)
    else: # ct
        for i in range(len(predictions)):
            item = predictions.iloc[i]
            write_ct(os.path.join(out_path, item.id +".ct"), item.id, item.sequence, item.base_pairs)
    


# imports
import os
import numpy as np
import torch as tr
import pandas as pd

from sincfold.embeddings import NT_DICT
from sincfold import __path__ as sincfold_path
import subprocess as sp 
from platform import system
import warnings

CT2DOT_CALL = f"export DATAPATH={sincfold_path[0]}/tools/RNAstructure/data_tables; {sincfold_path[0]}/tools/RNAstructure/ct2dot"
DRAW_CALL = f"export DATAPATH={sincfold_path[0]}/tools/RNAstructure/data_tables;  {sincfold_path[0]}/tools/RNAstructure/draw -c -u --svg -n 1"
VARNA_PATH = f"{sincfold_path[0]}/tools/varna/VARNAv3-93.jar"
if system() == "Windows":
    VARNA_PATH = ""
    CT2DOT_CALL = ""
    DRAW_CALL = ""

# All possible matching brackets for base pairing
MATCHING_BRACKETS = [
    ["(", ")"],
    ["[", "]"],
    ["{", "}"],
    ["<", ">"],
    ["A", "a"],
    ["B", "a"],
]
# Normalization.
BRACKET_DICT = {"!": "A", "?": "a", "C": "B", "D": "b"}


def pair_strength(pair):
    if "G" in pair and "C" in pair:
        return 3
    if "A" in pair and "U" in pair:
        return 2
    if "G" in pair and "U" in pair:
        return 0.8

    if pair[0] in NT_DICT and pair[1] in NT_DICT:
        n0, n1 = NT_DICT[pair[0]], NT_DICT[pair[1]]
        # Possible pairs with other bases
        if ("G" in n0 and "C" in n1) or ("C" in n0 and "G" in n1):
            return 3
        if ("A" in n0 and "U" in n1) or ("U" in n0 and "A" in n1):
            return 2
        if ("G" in n0 and "U" in n1) or ("U" in n0 and "G" in n1):
            return 0.8

    return 0


def prob_mat(seq):
    """Receive sequence and compute local conection probabilities (Ufold paper, optimized version)"""
    Kadd = 30
    window = 3
    N = len(seq)

    mat = np.zeros((N, N), dtype=np.float32)

    L = np.arange(N)
    pairs = np.array(np.meshgrid(L, L)).T.reshape(-1, 2)
    pairs = pairs[np.abs(pairs[:, 0] - pairs[:, 1]) > window, :]

    for i, j in pairs:
        coefficient = 0
        for add in range(Kadd):
            if (i - add >= 0) and (j + add < N):
                score = pair_strength((seq[i - add], seq[j + add]))
                if score == 0:
                    break
                else:
                    coefficient += score * np.exp(-0.5 * (add**2))
            else:
                break
        if coefficient > 0:
            for add in range(1, Kadd):
                if (i + add < N) and (j - add >= 0):
                    score = pair_strength((seq[i + add], seq[j - add]))
                    if score == 0:
                        break
                    else:
                        coefficient += score * np.exp(-0.5 * (add**2))
                else:
                    break

        mat[i, j] = coefficient

    return tr.tensor(mat)


def valid_mask(seq):
    """Create a NxN mask with valid canonic pairings."""

    seq = seq.upper().replace("T", "U")  # rna
    mask = tr.zeros((len(seq), len(seq)), dtype=tr.float32)
    for i in range(len(seq)):
        for j in range(len(seq)):
            if np.abs(i - j) > 3:  # nt that are too close are invalid
                if pair_strength([seq[i], seq[j]]) > 0:
                    mask[i, j] = 1
                    mask[j, i] = 1
    return mask


def normalize_brackets(struct):
    """Unify bracket notation"""
    for b in BRACKET_DICT:
        struct = struct.replace(b, BRACKET_DICT[b])
    return struct


def bracket_match(struct):
    match = True
    for pair in MATCHING_BRACKETS:
        match = match & (struct.count(pair[0]) == struct.count(pair[1]))
    return match


def fold2bp(struc, xop="(", xcl=")"):
    """Get base pairs from one page folding (using only one type of brackets).
    BP are 1-indexed"""
    openxs = []
    bps = []
    if struc.count(xop) != struc.count(xcl):
        return False
    for i, x in enumerate(struc):
        if x == xop:
            openxs.append(i)
        elif x == xcl:
            if len(openxs) > 0:
                bps.append([openxs.pop() + 1, i + 1])
            else:
                return False
    return bps


def dot2bp(struc):
    bp = []
    if not set(struc).issubset(
        set(["."] + [c for par in MATCHING_BRACKETS for c in par])
    ):
        return False

    for brackets in MATCHING_BRACKETS:
        if brackets[0] in struc:
            bpk = fold2bp(struc, brackets[0], brackets[1])
            if bpk:
                bp = bp + bpk
            else:
                return False
    return list(sorted(bp))


def dot2matrix(dot):
    matrix = tr.zeros((len(dot), len(dot)))
    base_pairs = dot2bp(dot)

    for bp in base_pairs:
        # base pairs are 1-based
        matrix[bp[0] - 1, bp[1] - 1] = 1
        matrix[bp[1] - 1, bp[0] - 1] = 1

    return matrix


def bp2matrix(L, base_pairs):
    matrix = tr.zeros((L, L))

    for bp in base_pairs:
        # base pairs are 1-based
        matrix[bp[0] - 1, bp[1] - 1] = 1
        matrix[bp[1] - 1, bp[0] - 1] = 1

    return matrix


def read_ct(ctfile):
    """Read ct file, return sequence and base_pairs"""
    seq, bp = [], []
    
    k = 1
    for line in open(ctfile):
        if line[0] == "#" or len(line.strip()) == 0:
            # comment
            continue

        line = line.split()
        if len(line) != 6 or not line[0].isnumeric() or not line[4].isnumeric:
            # header
            continue

        n1, n2 = int(line[0]), int(line[4])
        if k != n1: # add missing nucleotides as N
            seq += ["N"] * (n1-k)
        seq.append(line[1])
        k = len(seq) + 1
        if n2 > 0 and (n1 < n2):
            bp.append([n1, n2])
    return "".join(seq), bp


def write_ct(fname, seqid, seq, base_pairs):
    """Write ct file from sequence and base pairs. Base_pairs should be 1-based and unique per nt"""
    base_pairs_dict = {}
    for bp in base_pairs:
        base_pairs_dict[bp[0]] = bp[1]
        base_pairs_dict[bp[1]] = bp[0]

    with open(fname, "w") as fout:
        fout.write(f"{len(seq)} {seqid}\n")
        for k, n in enumerate(seq):
            fout.write(f"{k+1} {n} {k} {k+2} {base_pairs_dict.get(k+1, 0)} {k+1}\n")


def split_fasta_rec(s, mfe=True):
    mfe_start = s.rfind("(")

    if mfe:
        mfe = float(s[mfe_start + 1 : -1])
        s = s[:mfe_start]

    seq = s[: len(s) // 2]
    struct = s[len(s) // 2 :]
    return seq, struct, mfe


def mat2bp(x):
    """Get base-pairs from conection matrix [N, N]. It uses upper
    triangular matrix only, without the diagonal. Positions are 1-based"""
    ind = tr.triu_indices(x.shape[0], x.shape[1], offset=1)
    pairs_ind = tr.where(x[ind[0], ind[1]] > 0)[0]
    return (ind[:, pairs_ind].T + 1).tolist()


def postprocessing(preds, masks):
    """Postprocessing function using viable pairing mask.
    Inputs are batches of size [B, N, N]"""
    y_pred_mask = preds.multiply(masks)

    y_pred_mask_triu = tr.triu(y_pred_mask)
    y_pred_mask_max = tr.zeros_like(y_pred_mask)
    for k in range(y_pred_mask.shape[0]):
        y_pred_mask_max_aux = tr.zeros_like(y_pred_mask_triu[k, :, :])

        val, ind = y_pred_mask_triu[k, :, :].max(dim=0)
        y_pred_mask_max[k, ind[val > 0], val > 0] = val[val > 0]

        val, ind = y_pred_mask_max[k, :, :].max(dim=1)
        y_pred_mask_max_aux[val > 0, ind[val > 0]] = val[val > 0]

        ind = tr.where(y_pred_mask_max[k, :, :] != y_pred_mask_max_aux)
        y_pred_mask_max[k, ind[0], ind[1]] = 0

        y_pred_mask_max[k] = tr.triu(y_pred_mask_max[k]) + tr.triu(
            y_pred_mask_max[k]
        ).transpose(0, 1)
    return y_pred_mask_max

def find_pseudoknots(base_pairs):
    pseudoknots = []
    for i, j in base_pairs:
        for k, l in base_pairs:
            if i < k < j < l:  # pseudoknot definition
                if [k, l] not in pseudoknots:
                    pseudoknots.append([k, l])
    return pseudoknots

def dot2png(png_file, sequence, dotbracket, resolution=10):

    try:
        sp.run("java -version", shell=True, check=True, capture_output=True)
        sp.run(f'java -cp {VARNA_PATH} fr.orsay.lri.varna.applications.VARNAcmd -sequenceDBN {sequence} -structureDBN "{dotbracket}" -o  {png_file} -resolution {resolution}', shell=True)
    except:
        warnings.warn("Java Runtime Environment failed trying to run VARNA. Check if it is installed.")
    
    
def ct2svg(ct_file, svg_file):
    
    sp.run(f'{DRAW_CALL} {ct_file} {svg_file}', shell=True, capture_output=True)


def ct2dot(ct_file):
    if not os.path.isfile(ct_file) or os.path.splitext(ct_file)[1] != ".ct":
        raise ValueError("ct2dot requires a .ct file")
    dotbracket = ""
    if CT2DOT_CALL:
        sp.run(f"{CT2DOT_CALL} {ct_file} 1 tmp.dot", shell=True, capture_output=True)
        try: 
            dotbracket = open("tmp.dot").readlines()[2].strip()
            os.remove("tmp.dot")
        except FileNotFoundError: 
            print("Error in ct2dot: check .ct file")
    else:
        print("Dotbracket conversion only available on linux")
    return dotbracket

def bp2dot(L, base_pairs):
    """Create an inversible dot-bracket notation from base pairs"""
    dot = list("."*L)
    for i,j in base_pairs:
        dot[i-1] = "("
        dot[j-1] = ")"
    
    lvl = 1
    to_check = [base_pairs]
    while to_check:
        
        base_pairs = to_check.pop()
        
        pseudoknots = find_pseudoknots(base_pairs)
        if not pseudoknots:
            continue
        base_pairs = [bp for bp in base_pairs if bp not in pseudoknots]
        
        to_check += [base_pairs, pseudoknots]
    
        for i, j in pseudoknots:
            dot[i-1] = MATCHING_BRACKETS[lvl][0]
            dot[j-1] = MATCHING_BRACKETS[lvl][1]
        
        lvl += 1
        if lvl >= len(MATCHING_BRACKETS):
            break
    
    
    return "".join(dot)


from sincfold.embeddings import NT_DICT, VOCABULARY
def valid_sequence(seq):
    """Check if sequence is valid"""
    return set(seq.upper()) <= (set(NT_DICT.keys()).union(set(VOCABULARY)))

def validate_file(pred_file):
    """Validate input file fasta/csv format and return csv file"""
    if os.path.splitext(pred_file)[1] == ".fasta":
        table = []
        with open(pred_file) as f:
            row = [] # id, seq, (optionally) struct
            for line in f:
                if line.startswith(">"):
                    if row:
                        table.append(row)
                        row = []
                    row.append(line[1:].strip())
                else:
                    if len(row) == 1: # then is seq
                        row.append(line.strip())
                        if not valid_sequence(row[-1]):
                            raise ValueError(f"Sequence {row.upper()} contains invalid characters")
                    else: # struct
                        row.append(line.strip()[:len(row[1])]) # some fasta formats have extra information in the structure line
        if row:
            table.append(row)
        
        pred_file = pred_file.replace(".fasta", ".csv")
        
        if len(table[-1]) == 2:
            columns = ["id", "sequence"]
        else:
            columns = ["id", "sequence", "dotbracket"]

        pd.DataFrame(table, columns=columns).to_csv(pred_file, index=False)

    elif os.path.splitext(pred_file)[1] != ".csv":
        raise ValueError("Predicting from a file with format different from .csv or .fasta is not supported")
    
    return pred_file 


import pandas as pd
from torch.utils.data import Dataset
import torch as tr
import os
import json
import pickle
from sincfold.embeddings import OneHotEmbedding
from sincfold.utils import valid_mask, prob_mat, bp2matrix, dot2bp


class SeqDataset(Dataset):
    def __init__(
        self, dataset_path, min_len=0, max_len=512, verbose=False, cache_path=None, for_prediction=False, 
        use_restrictions=False, **kargs):
        self.max_len = max_len
        self.verbose = verbose
        if cache_path is not None and not os.path.isdir(cache_path):
            os.mkdir(cache_path)
        self.cache = cache_path

        # Loading dataset
        data = pd.read_csv(dataset_path)
        
        if for_prediction:
            assert (
                "sequence" in data.columns
                and "id" in data.columns
            ), "Dataset should contain 'id' and 'sequence' columns"

        else:
            assert (
                ("base_pairs" in data.columns or "dotbracket" in data.columns)
                and "sequence" in data.columns
                and "id" in data.columns
            ), "Dataset should contain 'id', 'sequence' and 'base_pairs' or 'dotbracket' columns"

            if "base_pairs" not in data.columns and "dotbracket" in data.columns:
                data["base_pairs"] = data.dotbracket.apply(lambda x: str(dot2bp(x)))      

        data["len"] = data.sequence.str.len()

        if max_len is None:
            max_len = max(data.len)
        self.max_len = max_len

        datalen = len(data)

        data = data[(data.len >= min_len) & (data.len <= max_len)]

        if len(data) < datalen:
            print(
                f"From {datalen} sequences, filtering {min_len} < len < {max_len} we have {len(data)} sequences"
            )

        self.sequences = data.sequence.tolist()
        self.ids = data.id.tolist()

        self.embedding = OneHotEmbedding()
        self.embedding_size = self.embedding.emb_size
        self.use_restrictions = use_restrictions

        self.base_pairs = None
        if "base_pairs" in data.columns:
            self.base_pairs = [
                json.loads(data.base_pairs.iloc[i]) for i in range(len(data))
            ]

    def __len__(self):
        return len(self.sequences)

    def __getitem__(self, idx):
        seqid = self.ids[idx]
        cache = f"{self.cache}/{seqid}.pk"
        if (self.cache is not None) and os.path.isfile(cache):
            item = pickle.load(open(cache, "rb"))
        else:
            sequence = self.sequences[idx]
            L = len(sequence)
            Mc = None
            if self.base_pairs is not None:
                Mc = bp2matrix(L, self.base_pairs[idx])

            seq_emb = self.embedding.seq2emb(sequence)

            mask = valid_mask(sequence)
            if self.use_restrictions:
                prob_mask = prob_mat(sequence)
                item = [seq_emb, Mc, L, mask, seqid, sequence, prob_mask]
            else:
                item = [seq_emb, Mc, L, mask, seqid, sequence]
                

            if self.cache is not None:
                pickle.dump(item, open(cache, "wb"))
                
        return item


def pad_batch(batch):
    """batch is a list of (seq_emb, Mc, L, mask, prob_mask, seqid)"""
    if len(batch[0]) == 7:
        seq_emb, Mc, L, mask, seqid, sequence, prob_mask = zip(*batch)
    else:
        prob_mask = None
        seq_emb, Mc, L, mask, seqid, sequence = zip(*batch)

    seq_emb_pad = tr.zeros((len(batch), seq_emb[0].shape[0], max(L)))
    if Mc[0] is None:
        Mc_pad = None
    else:
        Mc_pad = -tr.ones((len(batch), max(L), max(L)), dtype=tr.long)
    mask_pad = tr.zeros((len(batch), max(L), max(L)))
    if prob_mask is not None:
        prob_mask_pad = tr.zeros((len(batch), max(L), max(L)))

    for k in range(len(batch)):
        seq_emb_pad[k, :, : L[k]] = seq_emb[k]
        if Mc_pad is not None:
            Mc_pad[k, : L[k], : L[k]] = Mc[k]
        mask_pad[k, : L[k], : L[k]] = mask[k]
        if prob_mask is not None:
            prob_mask_pad[k, : L[k], : L[k]] = prob_mask[k]

    if prob_mask is not None:
        return seq_emb_pad, Mc_pad, L, mask_pad, seqid, sequence, prob_mask_pad
    else:
        return seq_emb_pad, Mc_pad, L, mask_pad, seqid, sequence
